from django.db import models
from django.contrib.auth.models import Group, Permission
from django.contrib.auth.models import AbstractUser


class User(AbstractUser):
    full_name = models.CharField(max_length=256, blank=True)
    phone = models.CharField(max_length=256, blank=True, null=True)

    @property
    def is_admin(self):
        return super().is_staff or self.groups.filter(name='admin').exists()

    @property
    def is_student(self):
        return self.groups.filter(name='student').exists()

    @property
    def is_teacher(self):
        return self.groups.filter(name='teacher').exists()


__title__ = 'DRF Guard'
__description__ = 'Flexible permissions for Django REST Framework'
__url__ = 'https://github.com/yezyilomo/drf-guard'
__version__ = '0.1.1'
__author__ = 'Yezy Ilomo'
__author_email__ = 'yezileliilomo@hotmail.com'
__license__ = 'MIT'
__copyright__ = 'Copyright 2020 Yezy Ilomo'

# Version synonym
VERSION = __version__



from rest_framework import permissions
from django.contrib.auth.models import Group, Permission

from .operators import Operator, Reducer


class HasRequiredGroups(permissions.BasePermission):
    """
    Ensure user is in required groups.
    """
    @classmethod
    def is_in_group(cls, user, group):
        if group == '__all__':
            return True
        if isinstance(group, (str, Group)):
            return user.groups.filter(name=group).exists()
        if isinstance(group, (list, tuple)):
            return cls.is_in_required_groups(user, group)
        if issubclass(group, Operator):
            return group()

    @classmethod
    def is_in_required_groups(cls, user, groups):
        if groups is None:
            # Don't check the permissions
            return True
            
        reducer = Reducer()
        return reducer(
                [cls.is_in_group(user, group) for group in groups]
            )

    def has_permission(self, request, view):
        if view.action == 'retrieve':
            # Separate retrive URL(This will be handled in has_object_permission)
            return True

        # Get a mapping of methods -> required group.
        required_groups_mapping = getattr(view, "groups_and_permissions", {})

        # Determine the required groups for this particular request method.
        required_groups = required_groups_mapping.get(request.method, {})

        if view.action == 'list':
            # Get required groups for list action
            required_groups = required_groups.get('list', {'groups': None, 'permissions': None})
        
        required_groups = required_groups.get('groups', None)

        # Return True if the user has all the required groups or is staff.
        return self.is_in_required_groups(request.user, required_groups)

    def has_object_permission(self, request, view, obj):
        if view.action == 'list':
            # Separate list URL(This will be handled in has_permission)
            return True

        # Get a mapping of methods -> required group.
        required_groups_mapping = getattr(view, "groups_and_permissions", {})

        # Determine the required groups for this particular request method.
        required_groups = required_groups_mapping.get(request.method, {})

        if view.action == 'retrieve':
            # Get required groups for retrieve action
            required_groups = required_groups.get('retrieve', {'groups': None, 'permissions': None})

        required_groups = required_groups.get('groups', None)

        # Return True if the user has all the required groups or is staff.
        return self.is_in_required_groups(request.user, required_groups)


class HasRequiredPermissions(permissions.BasePermission):
    """
    Ensure user is has required permissions.
    """
    @classmethod
    def has_required_permission(cls, permission, request, view, obj=None):
        if isinstance(permission, str):
            return request.user.has_perm(permission)
        elif isinstance(permission, (list, tuple)):
            return cls.has_required_permissions(permission, request, view, obj)
        elif isinstance(permission, Permission):
            return request.user.has_perm(permission.codename)
        elif issubclass(permission, permissions.BasePermission):
            if obj is None:
                return permission().has_permission(request, view)
            else:
                return permission().has_object_permission(request, view, obj)
        elif issubclass(permission, Operator):
            # Encountered an operator
            return permission()
        else:
            data_type = type(permission).__name__
            raise TypeError("`%s` is invalid permission type." % data_type)

    @classmethod
    def has_required_permissions(cls, permissions, *args):
        if permissions is None:
            # Don't check the permissions
            return True

        reducer = Reducer()
        return reducer([
            cls.has_required_permission(permission, *args)
            for permission in permissions
        ])
        
    @staticmethod
    def get_permissions(request, view):
        # Get a mapping of methods -> required group.
        required_permissions_mapping = getattr(view, "groups_and_permissions", {})

        # Determine the required groups for this particular request method.
        required_permissions = required_permissions_mapping.get(request.method, {})

        if view.action == 'retrieve':
            required_permissions = required_permissions.get('retrieve', {'groups': None, 'permissions': None})
        elif view.action == 'list':
            required_permissions = required_permissions.get('list', {'groups': None, 'permissions': None})

        return required_permissions.get('permissions', None)

    def has_permission(self, request, view):
        required_permissions = self.get_permissions(request, view)
        return self.has_required_permissions(required_permissions, request, view)

    def has_object_permission(self, request, view, obj):
        required_permissions = self.get_permissions(request, view)
        return self.has_required_permissions(required_permissions, request, view, obj)



from functools import reduce


class Operator():
    @property
    def should_eval(self):
        return hasattr(self, 'left_operand') and hasattr(self, 'right_operand')


class Or(Operator):
    @property
    def should_eval(self):
        if getattr(self, 'left_operand', False):
            # Shortcircuit
            raise StopIteration(self.left_operand)
        return super().should_eval

    def eval(self):
        return self.left_operand or self.right_operand


class And(Operator):
    @property
    def should_eval(self):
        if not getattr(self, 'left_operand', True):
            # Shortcircuit
            raise StopIteration(self.left_operand)
        return super().should_eval

    def eval(self):
        return self.left_operand and self.right_operand


class Not(Operator):
    @property
    def should_eval(self):
        return hasattr(self, 'right_operand')
        
    def eval(self):
        if hasattr(self, 'left_operand') and isinstance(self.left_operand, Operator):
            # This is the case of left-And-Not-right / left-Or-Not-right
            self.left_operand.right_operand = not self.right_operand
            return self.left_operand.eval()
        return not self.right_operand
    

class Reducer():
    def __call__(self, sequence):
        return self.eval(sequence)

    def reducer(self, left_operand, right_operand):
        if isinstance(left_operand, Operator) and isinstance(right_operand, Operator):
            right_operand.left_operand = left_operand
            return right_operand
            
        if isinstance(right_operand, (list, tuple)):
            right_operand = reduce(self.reducer, right_operand)
            
        if isinstance(left_operand, (list, tuple)):
            left_operand = reduce(self.reducer, left_operand)
            
        if isinstance(left_operand, Operator):
            left_operand.right_operand = right_operand
            if left_operand.should_eval:
                return left_operand.eval()
            return left_operand
            
        if isinstance(right_operand, Operator):
            right_operand.left_operand = left_operand
            if right_operand.should_eval:
                return right_operand.eval()
            return right_operand
            
        return left_operand and right_operand # Default operator for ','
    
    def eval(self, sequence):
        try:
            return reduce(self.reducer, sequence)
        except StopIteration as e:
            # Return value due to shortcircuit
            return e.args[0]




from rest_framework import viewsets
from rest_framework.permissions import IsAuthenticated

from tests.testapp.models import User
from tests.testapp.serializers import UserSerializer

from drf_guard.operators import And, Or, Not
from .permissions import IsAdminUser, IsAllowedUser, IsTeacherAccessingStudent
from drf_guard.permissions import HasRequiredGroups, HasRequiredPermissions


class UserViewSet(viewsets.ModelViewSet):
    """API endpoint that allows users to be viewed or edited."""
    queryset = User.objects.all()
    serializer_class = UserSerializer
    permission_classes = (HasRequiredGroups, HasRequiredPermissions)
    http_method_names = ['get', 'put', 'patch', 'delete']
    groups_and_permissions = {
         'GET': {
             'list': {
                 'groups': ['admin'],
                 'permissions': [IsAuthenticated]
             },
             'retrieve': {
                 'groups': ['__all__'],
                 'permissions': [
					 IsAuthenticated, And, 
					 [IsAllowedUser, Or, IsAdminUser, Or, IsTeacherAccessingStudent]
					]
             },
         },
         'PUT': {
             'groups': ['__all__'],
             'permissions': [IsAuthenticated, [IsAllowedUser, Or, IsAdminUser]]
         },
         'PATCH': {
             'groups': ['__all__'],
             'permissions': [IsAuthenticated, [IsAllowedUser, Or, IsAdminUser]]
         },
         'DELETE': {
             'groups': ['__all__'],
             'permissions': [IsAuthenticated, [IsAllowedUser, Or, IsAdminUser]]
         }
    }



from rest_framework import permissions


class IsAllowedUser(permissions.BasePermission):
    """
    Custom permission to only allow owners of an object to edit it.
    """

    def has_object_permission(self, request, view, obj):
        return obj == request.user


class IsAdminUser(permissions.BasePermission):
    """
    Custom permission to check if user is admin
    """

    def has_permission(self, request, view):
        return request.user.is_authenticated and request.user.is_admin

    def has_object_permission(self, request, view, obj):
        return request.user.is_authenticated and request.user.is_admin


class IsTeacherAccessingStudent(permissions.BasePermission):
    """
    Custom permission to check if user is admin
    """

    def has_object_permission(self, request, view, obj):
        return request.user.is_teacher and obj.is_student


import os
import re
import sys

from codecs import open
from setuptools import setup, find_packages


# 'setup.py publish' shortcut.
if sys.argv[-1] == 'publish':
    os.system('python3 setup.py sdist bdist_wheel')
    os.system('twine upload dist/*')
    sys.exit()

def get_readme():
    readme = ''
    with open('README.md', 'r', 'utf-8') as f:
        readme = f.read()
    return readme

def get_info(info_name):
    init_py = open(os.path.join('drf_guard', '__init__.py')).read()
    return re.search("%s = ['\"]([^'\"]+)['\"]" % info_name, init_py).group(1)

url = get_info('__url__')
version = get_info('__version__')
license_ = get_info('__license__')
description = get_info('__description__')
author = get_info('__author__')
author_email = get_info('__author_email__')
readme = get_readme()

setup(
    name = 'drf-guard',
    version = version,
    description = description,
    long_description = readme,
    long_description_content_type = 'text/markdown',
    url = url,
    author = author,
    author_email = author_email,
    license = license_,
    packages = find_packages(exclude=('tests','test')),
    package_data = {'': ['LICENSE']},
    install_requires = [
        'django>=1.11',
        'djangorestframework>=3.5'
    ],
    python_requires = '>=3.5',
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Framework :: Django',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.0',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ]
)



from rest_framework import serializers
from tests.testapp.models import User


ROLE_CHOICES = (
    ('admin', 'Admin'),
    ('student', 'Student'),
    ('teacher', 'Teacher')
)

class UserSerializer(serializers.ModelSerializer):
    class Meta:
        model = User
        fields = (
            'id', 'username'
        )
